/* Camera Implementation - Camera Component - Model Branch
*
* Camera.cpp
* fwk4gps version 1.0
* gam666/dps901
* August 30 2010
* copyright (c) 2010 Chris Szalwinski 
* distributed under TPL - see ../Licenses.txt
*/

#include "iContext.h"        // for the Context Interface
#include "iViewing.h"        // for the Viewing Interface
#include "iUtilities.h"      // for error(), sprintf()

#include "Camera.h"          // for Camera class declaration
#include "MathDefinitions.h" // for math functions in model coordinates
#include "ModelSettings.h"   // for FLOOR, FORWARD_SPEED, ROT_SPEED, MAX_DESC
#include "GraphicsPlatform.h"
#include "Configuration.h"
#include <math.h>

//-------------------------------- Camera -------------------------------------
//
// The Camera Primitive represents one viewpoint in the Viewing Component
//
// The Viewing Coordinator manages the Camera Instances
//
iViewing* Camera::coordinator = NULL;

// CreateCamera creates a Camera object, which implements the Camera Primitive
//
iCamera* CreateCamera(iContext* c) {

    return new Camera(c);
}

// constructor adds the Camera Instance to the Viewing Coordinator
//
Camera::Camera(iContext* c) : context(c), host(NULL) {

    if (!coordinator)
        error(L"Camera::00 Couldn\'t access the Viewing Coordinator");
    else if(!coordinator->add(this))
        error(L"Camera::01 Couldn\'t add the camera to the Viewing Coordinator");
    rxl=0.0f;
    ryl=0.0f;
#if MODEL_Z_AXIS == FAR_TO_NEAR
    rotatey(3.14159f);
#endif
}

// attach attaches the Camera Instance to host Frame f
//
void Camera::attach(const iFrame* f) {

    host = f;
}

// update adjusts the camera's viewpoint and heading according
// to the user's interventions
//
void Camera::update(int delta) {

	if(!context->move_locked()){//locks camera movement if true
    int dx = 0, // pitch up/down
        dy = 0, // yaw left/right
        dz = 0; // advance/retreat
    int mouseLeftRight=0;
    int mouseUpDown =0;

    // keyboard input
    if (context->pressed(MOVE_UP))       dz += delta*25;
    if (context->pressed(MOVE_DOWN))     dz -= delta*25;
    if (context->pressed(MOVE_LEFT))     dy -= delta*25;
    if (context->pressed(MOVE_RIGHT))    dy += delta*25;
    
    Vector right = Vector(1.0f,0.0f,0.0f);//world right vector
    Vector up = Vector(0.0f,1.0f,0.0f);//world up vector
    Vector direction = Vector(0.0f,0.0f,1.0f);//world direction vector    
    float ymouse = context->get(GF_MS_POSY);//gets the mouse Y offset
    float xmouse = context->get(GF_MS_POSX);//gets the mouse X offset
    float rzl=0.0f;
    rxl += ((ToRadians(ymouse)) * ROT_SPEED*225 ); //set rxl to the radian value of ymouse times the rotation speed
    ryl += ((ToRadians(xmouse)) * ROT_SPEED*225);  //set ryl to the radian value of xmouse times the rotation speed

      // apply the translations and rotations
     if (rxl || ryl) {
        // Rotate the Camera
        Matrix rot; // rotation matrix
		// rotate about the up axis (yaw)
        rot        = rotate(up, ryl);
        direction *= rot;
        right     *= rot;
        // rotate about the right axis (pitch)
        
        rot        = rotate(right, rxl);
        direction *= rot;
        up        *= rot;
        
		 // rotate about the heading (roll)
        rot        = rotate(direction, 0.0f);
        up        *= rot;
        right     *= rot;
        
    }
    // adjust camera vectors
    if (dx || dy || dz) {
        //Move to the left/right without changing direction
        if (dy){
            Vector heading = right;
            Vector displays = (dy * (FORWARD_SPEED)) * heading;
            translate(displays.x,0.0f,displays.z * MODEL_Z_AXIS);

        }
        //Move forward/backwards without changing direction
        if (dz) { 
            Vector heading = direction;
            Vector disp = (dz * FORWARD_SPEED) * heading; 
            translate(disp.x, 0.0f, disp.z * MODEL_Z_AXIS);
            // bottom out when we hit the floor
            Vector p = position();
            if (p.y <= FLOOR) {
                translate(0, FLOOR - p.y, 0);
                heading = normal(heading);
                if (heading.y < 0) {
                    float rot = atanf(heading.y);
                    rotatex(rot);
                }
            }
        }
    }

    // store the current viewpoint, heading and up direction
    Vector p = host ? position() * host->world() : position();
    context->set(GF_CA_POSN, p);//sets the camera position in the context object
    context->set(GF_CA_HEAD, direction);//sets the camera direction in the context object
    context->set(GF_CA_UP,   up);//sets the camera up in the context object
	}
}

// destructor removes the Camera Instance from the Viewing Coordinator
//
Camera::~Camera() {

    if (coordinator)
        coordinator->remove(this);
    else
        error(L"Camera::90 Couldn\'t access the Viewing Coordinator");
}

// position returns the text string with the camera's position
//
const wchar_t* position(wchar_t* str, const iCamera* camera) {

    if (camera) {
        Vector p = camera->position();
        p.z *= Z_AXIS;
        sprintf(str, (int)p.x, (int)p.y, (int)p.z);
    }

    return str;
}

// heading returns the text string with the camera's heading
//
const wchar_t* heading(wchar_t* str, const iCamera* camera) {

    if (camera) {
        Vector h = camera->orientation('z');
        h.z *= Z_AXIS;
        sprintf(str, (int)(100 * h.x), (int)(100 * h.y), (int)(100 * h.z));
    }

    return str;
}
