/* Display Implementation - GraphicsCard Component - Direct Device Branch
 *
 * Display.cpp
 * fwk4gps version 1.0
 * gam666/dps901
 * August 30 2010
 * copyright (c) 2010 Chris Szalwinski 
 * distributed under TPL - see ../Licenses.txt
 */

#include "iContext.h"        // for the Context Interface
#include "iUtilities.h"      // for error()
#include "Display.h"         // for the Display class declaration
#include "DeviceLight.h"     // for access to DeviceLight::connectsThrough()
#include "Graphic.h"         // for access to Graphic::connectsThrough()
#include "DeviceTexture.h"   // for access to DeviceTexture::connectsThrough()
#include "Font.h"            // for access to Font::connectsThrough()
#include "MathDefinitions.h" // for Matrix, Vector and Colour
#include "Configuration.h"   // for MAX_DESC, projection, background settings  

//-------------------------------- Display ------------------------------------
//
// The Display Portal manages the API connectivity for the Graphics Card 
// Component
//
// CreateDisplay creates the Display Object, which implements the Display Portal
//
iDisplay* CreateDisplay(iContext* c) {

	return new Display(c);
}

// constructor initializes the instance variables and aligns the Display Portal
// with the other Elements of the Graphics Card Component
//
Display::Display(iContext* c) : context(c), hwnd(NULL), width_(0), height_(0), 
 lightIndex(NULL) {

    // obtain Interface to Direct3D COM object
	d3d = Direct3DCreate9(D3D_SDK_VERSION);
    if (!d3d)
        error(L"Display::01 Failed to make Direct3D object");
    d3dd           = NULL;
    spriteManager_ = NULL;
    modeDim        = NULL;

	// the Display Portal provides the API connectivity for the following Elements
	DeviceLight::connectsThrough(this);
	Graphic::connectsThrough(this);
	DeviceTexture::connectsThrough(this);
	Font::connectsThrough(this);
}

// interrogate the host system for the available display configurations
//
bool Display::interrogate(void* hwnd) {

    bool     rc = false;
    int      nAdapters, nModes, nPixelFormats;
	wchar_t  str[MAX_DESC + 1];

    D3DADAPTER_IDENTIFIER9 d3di;
    D3DDISPLAYMODE         mode;
    D3DFORMAT              Format[]  = D3D_DOC_FORMATS;
    // friendly descriptions for each format
    wchar_t*               fmtdesc[] = D3D_FORMAT_DESC;

    // adapter count
    nAdapters = d3d->GetAdapterCount();
    // maximum mode count
    nModes = 0;
    // pixel format count
    // number of pixel formats as described in the Direct3D documentation
    nPixelFormats = D3D_NO_DOC_FORMATS;
 
    // determine maximum number of modes with any adapter
    for (int id = 0; id < nAdapters; id++) {
        if (SUCCEEDED(d3d->GetAdapterIdentifier(id, 0, &d3di))) {
            for (int ip = 0; ip < nPixelFormats; ip++) {
                int i = d3d->GetAdapterModeCount(id, Format[ip]);
                if (i > nModes)
                    nModes = i;
            }
        }
    }
    if (nAdapters > MAX_ADAPTERS) {
        wchar_t str[MAX_DESC + 1];
        sprintf(str, nAdapters, L" Adapters found - increase MAX_ADAPTERS");
        error(str);
        nAdapters = MAX_ADAPTERS;
    }
    if (nModes > MAX_MODES) {
        wchar_t str[MAX_DESC + 1];
        sprintf(str, nModes, L" Modes found - increase MAX_MODES");
        error(str);
        nModes = MAX_MODES;
    }
    if (nPixelFormats > MAX_P_FORMATS) {
        wchar_t str[MAX_DESC + 1];
        sprintf(str, nPixelFormats, L" Pixel Formats found - increase MAX_P_FORMATS");
        error(str);
        nPixelFormats = MAX_P_FORMATS;
    }

    // set dimensions, allocate context memory for descriptions,
    // and allocate private memory for mode dimensions
    context->set(GF_DS_ACNT, nAdapters);
    context->set(GF_DS_MCNT, nModes);
    context->set(GF_DS_PCNT, nPixelFormats);
    if (modeDim) delete [] modeDim;
    modeDim = new int[nAdapters * nModes * nPixelFormats][2];

    // enumerate and set all descriptions
    for (int id = 0; id < nAdapters; id++) {
        if (SUCCEEDED(d3d->GetAdapterIdentifier(id, 0, &d3di))) {
            rc = false;
            for (int ip = 0; ip < nPixelFormats; ip++) {
                // mode count
                int nr = d3d->GetAdapterModeCount(id, Format[ip]);
                nr = (nr > nModes) ? nModes : nr;
                for (int ir = 0; ir < nr; ir++) {
                    if (SUCCEEDED(d3d->EnumAdapterModes(id, Format[ip], ir, &mode))
                     && mode.Width >= WND_WIDTH && mode.Height >= WND_HEIGHT &&
                     (D3D_OK == d3d->CheckDeviceFormat(id, D3DDEVTYPE_HAL, mode.Format,
                     D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE,  D3DFMT_D32) ||   
                     D3D_OK == d3d->CheckDeviceFormat(id, D3DDEVTYPE_HAL, mode.Format,
                     D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D16) ||
                     D3D_OK == d3d->CheckDeviceFormat(id, D3DDEVTYPE_REF, mode.Format,
                     D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D32) ||
                     D3D_OK == d3d->CheckDeviceFormat(id, D3DDEVTYPE_REF, mode.Format,
                     D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D16))) {
                        wchar_t hz[20] = L"";
                        if (mode.RefreshRate)
                            wsprintf(hz, L"(%d Hz)", mode.RefreshRate);
                        wsprintf(str, L"%dx%d %ls %ls bits", mode.Width, mode.Height,
                         hz, fmtdesc[ip]);
                        rc = true;
                        // store the description and dimensions
                        int i = (id * nModes + ir) * nPixelFormats + ip;
                        context->set(GF_MDEDESC,   i, str);
                        modeDim[i][0] = mode.Width;
                        modeDim[i][1] = mode.Height;
                    }
                }
            }
            if (rc) {
                // adapter description
		        strcpyFromMB(str, d3di.Description, MAX_DESC); 
                context->set(GF_ADPDESC, id, str);
            }
        }
    }

    return rc;
}

// setConfiguration sets the configuration to be used for setup
//
void Display::setConfiguration() {

    display     = context->get(GF_DS_ADPT);
    mode        = context->get(GF_DS_MODE);
    pixel       = context->get(GF_DS_PIXL);
    runInWindow = context->get(GF_DS_RINW) == 1;
    if (runInWindow) {
        width_  = WND_WIDTH;
        height_ = WND_HEIGHT;
    }
    else {
        int nModes  = context->get(GF_DS_MCNT);
        int nPixls  = context->get(GF_DS_PCNT);
        int i       = (display * nModes + mode) * nPixls + pixel;
        width_      = modeDim[i][0];
        height_     = modeDim[i][1];
    }
    context->set(GF_DS_WDTH, width_);
    context->set(GF_DS_HGHT, height_);
}

// setup creates the display device according to the context
// parameters and associates the device with the application window
// (HWND)hwnd
//
wchar_t* MultiCharToUniChar(char* mbString)//helper function that converts strings to unicode characters
{
	int len = 0;
	len = (int)strlen(mbString) + 1;
	wchar_t *ucString = new wchar_t[len];
	mbstowcs(ucString, mbString, len);
	return ucString;
}
bool Display::setup(void* hwnd) {

    bool rc = false;

	// store address of main application window
    this->hwnd = (HWND)hwnd;
	
			

    // set the D3D presentation parameters
    UINT adapter;
	D3DFORMAT d3dFormat;
    ZeroMemory(&d3dpp, sizeof d3dpp);
    d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_ONE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferCount = 1;
    d3dpp.EnableAutoDepthStencil = TRUE;
    D3DDISPLAYMODE d3ddm;
    if (!runInWindow) {
		D3DFORMAT Format[] = D3D_DOC_FORMATS;
		d3dFormat = Format[pixel];
        if (FAILED(d3d->EnumAdapterModes(display, d3dFormat, mode, &d3ddm))) {
            error(L"Display::10 Failed to get selected display mode");
            error(L"Display::11 Defaulting to windowed mode");
            runInWindow = true;
        }
        else {
            adapter = display;
            d3dpp.BackBufferWidth  = d3ddm.Width;
            d3dpp.BackBufferHeight = d3ddm.Height;
            d3dpp.BackBufferFormat = d3ddm.Format;
            d3dpp.FullScreen_RefreshRateInHz = d3ddm.RefreshRate;
            width_  = d3ddm.Width;
            height_ = d3ddm.Height;
        }
    }
    if (runInWindow) {
        adapter = D3DADAPTER_DEFAULT;
		d3d->GetAdapterDisplayMode(adapter, &d3ddm);
		d3dFormat = d3ddm.Format;
        d3dpp.Windowed = TRUE;
		d3dpp.BackBufferWidth  = WND_WIDTH;
		d3dpp.BackBufferHeight = WND_HEIGHT;
		d3dpp.BackBufferFormat = d3dFormat;
        width_  = WND_WIDTH;
        height_ = WND_HEIGHT;
    }
	aspect = (float) width_ / height_;

    // find the best format for depth buffering and stenciling
    //
    D3DDEVTYPE devtype;
    if (D3D_OK == d3d->CheckDeviceFormat(adapter, D3DDEVTYPE_HAL,
     d3dFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D32)) {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D32; // depth buffer
        devtype = D3DDEVTYPE_HAL;                  // HAL device   
    }
    else if (D3D_OK == d3d->CheckDeviceFormat(adapter, D3DDEVTYPE_HAL, 
     d3dFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D16)) {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D16;  // depth buffer
        devtype = D3DDEVTYPE_HAL;                   // HAL Device
    }
    // if the above attempts fail, use the REF (software emulation) device
    // with a 32-bit depth buffer rather than the HAL (hardware accelerated) 
	// device
    else if (D3D_OK == d3d->CheckDeviceFormat(adapter, D3DDEVTYPE_REF,
     d3dFormat, D3DUSAGE_DEPTHSTENCIL, D3DRTYPE_SURFACE, D3DFMT_D32)) {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D32;   // depth buffer
        devtype = D3DDEVTYPE_REF;                    // REF Device
    }
    // if all else fails, use the REF (software emulation) with a 16-bit
    // depth buffer, hoping that it will work. (If it doesn't, we are out
    // of luck anyway.)
    else {
        d3dpp.AutoDepthStencilFormat = D3DFMT_D16;   // depth buffer
        devtype = D3DDEVTYPE_REF;                    // REF Device
    }

    // extract the device capabilities and configure the limits
    D3DCAPS9 caps;
    d3d->GetDeviceCaps(adapter, devtype, &caps);
	DWORD behaviorFlags;
	// hardware or software vertex processing?
	if ((caps.DevCaps & D3DDEVCAPS_HWTRANSFORMANDLIGHT) == 0 ||
	 caps.VertexShaderVersion < D3DVS_VERSION(1,1))
	    behaviorFlags = D3DCREATE_SOFTWARE_VERTEXPROCESSING;
	else
	    behaviorFlags = D3DCREATE_HARDWARE_VERTEXPROCESSING;
    // retrieve the Interface to the D3D display device
    if (d3dd)
        error(L"Display::11 Pointer to Direct3D interface is not NULL");
    else if (FAILED(d3d->CreateDevice(adapter, devtype, (HWND)hwnd, 
	 behaviorFlags, &d3dpp, &d3dd)))
        error(L"Display::12 Failed to create Direct3D device");
    else {
        // maximum number of lights supported by the display device
		maxLights = caps.MaxActiveLights ? caps.MaxActiveLights : 8;
        // set anisotropic filtering to the maximum available on the device
        if (FAILED(d3dd->SetSamplerState(0, D3DSAMP_MAXANISOTROPY,
         caps.MaxAnisotropy - 1)))
            error(L"Display::13 Failed to set up anisotropic filtering");

		// create a sprite COM object to manage the drawing of the hud texture
		// and the drawing of the text item fonts
		if (!spriteManager_ && FAILED(D3DXCreateSprite(d3dd, &spriteManager_)))
			error(L"Display::14 Failed to create the font manager");
		if(d3dd){
			LPD3DXBUFFER materialBuffer=NULL;
            numMaterials=0;            // Note: DWORD is a typedef for unsigned long
             mesh=NULL;
 // Load the mesh from the specified file
            HRESULT hr=D3DXLoadMeshFromX(L"gam666_levelv8.X", D3DXMESH_SYSTEMMEM,
                                            d3dd, NULL,
                                         &materialBuffer,NULL, &numMaterials,
                                         &mesh );
			int test;
			if(hr != D3D_OK)   
    {   
        switch(hr)   
        {   
        case D3DERR_INVALIDCALL:   
            //MessageBox(NULL,"D3DERR_INVALIDCALL from D3DXLoadMeshFromX() in Model::LoadFromX()","Error",MB_OK | MB_ICONERROR); 
			test = 1;
            break;   
        case E_OUTOFMEMORY:   
           // MessageBox(NULL,"E_OUTOFMEMORY from D3DXLoadMeshFromX() in Model::LoadFromX()","Error",MB_OK | MB_ICONERROR);   
			test = 2;
            break;   
        case D3DXFERR_FILENOTFOUND:   
           // MessageBox(NULL,"D3DXFERR_FILENOTFOUND from D3DXLoadMeshFromX() in Model::LoadFromX()","Error",MB_OK | MB_ICONERROR);  
			test=3;
            break;   
       // default:   
       // {   
            //char error[1000];   
            //sprintf(error,"Unknown error from D3DXLoadMeshFromX() in Model::LoadFromX(): %d",result);   
            //MessageBox(NULL,error,"Error",MB_OK | MB_ICONERROR);   
       // }   
        }  
			}


 D3DXMATERIAL* d3dxMaterials = (D3DXMATERIAL*)materialBuffer->GetBufferPointer();
             meshMaterials = new D3DMATERIAL9[numMaterials];
            meshTextures  = new LPDIRECT3DTEXTURE9[numMaterials];
            for (DWORD i=0; i<numMaterials; i++)
            {
           // Copy the material
 meshMaterials[i] = d3dxMaterials[i].MatD3D;

 // Set the ambient color for the material (D3DX does not do this)
 meshMaterials[i].Ambient = meshMaterials[i].Diffuse;
     
 // Create the texture if it exists - it may not
 meshTextures[i] = NULL;
 if (d3dxMaterials[i].pTextureFilename && meshTextures[i] == NULL){
	 wchar_t* tex_temp = MultiCharToUniChar(d3dxMaterials[i].pTextureFilename);
	  D3DXCreateTextureFromFileEx(d3dd, tex_temp, D3DX_DEFAULT,
      D3DX_DEFAULT, D3DX_DEFAULT, NULL, D3DFMT_A8R8G8B8, D3DPOOL_DEFAULT,
      D3DX_DEFAULT, D3DX_DEFAULT, D3DCOLOR_XRGB(0, 0, 0), NULL, NULL,
      &meshTextures[i]); 
	 delete tex_temp;
 }

			}

		}
		// setup successful
        rc = true;
    }

	// complete the setup
	if (rc) {
        setupProjection();
        setupLighting();
		setupBlending();
	}

    return rc;
}

// setupProjection calculates the transformation from camera space to
// homogeneous clip space and applies the transformation to the display
// device
//
void Display::setupProjection() {

	Matrix projection = ::projection(FIELD_OF_VIEW, aspect, NEAR_CLIPPING, 
     FAR_CLIPPING);

	if (d3dd)
		d3dd->SetTransform(D3DTS_PROJECTION, (D3DXMATRIX*)&projection);
}

// setupLighting sets the lighting parameters on the graphics card
//
void Display::setupLighting() {

    // allow specular highlights (can be slow on some machines)
    if (d3dd) d3dd->SetRenderState(D3DRS_SPECULARENABLE, TRUE);

    if (lightIndex)
        delete [] lightIndex;
    lightIndex = new bool[maxLights];
    for (int i = 0; i < maxLights; i++)
        lightIndex[i] = true;
}

// setupBlending sets up colour dithering and sets the formula for
// alpha blending
//
void Display::setupBlending() {

    if (d3dd) {
        // allow colour dithering (much smoother looking when using lights)
        d3dd->SetRenderState(D3DRS_DITHERENABLE, TRUE);

        // how alpha-blending is done (when drawing transparent things)
        d3dd->SetRenderState(D3DRS_SRCBLEND, D3DBLEND_SRCALPHA);
        d3dd->SetRenderState(D3DRS_DESTBLEND, D3DBLEND_INVSRCALPHA);
    }
}

// getLightIndex gets the next light index on the display device
//
int Display::getLightIndex() {

    int rc = -1;

    if (lightIndex)
        for (int i = 0; i < maxLights; i++)
            if (lightIndex[i]) {
                lightIndex[i] = false;
                rc = i;
                i  = maxLights;
            }

    return rc;
}

// restore re-activates the display device
//
bool Display::restore() {

    bool rc = false;

    if (d3dd) {
        HRESULT hr;
		hr = d3dd->TestCooperativeLevel();
		if (hr == D3DERR_DEVICENOTRESET)
			// reset the display device
			rc = d3dd->Reset(&d3dpp) == D3D_OK;
		else if (hr == S_OK)
			rc = true;
	}
	if (rc) {
		// reacquire sprite manager references to video memory
		if (spriteManager_) spriteManager_->OnResetDevice();
    }

	// complete the restoration
	if (rc) {
        setupProjection();
        setupLighting();
		setupBlending();
	}

    return rc;
}

// beginDraw applies the view transformation, sets the ambient lighting, and
// clears the backbuffer
//



void Display::beginDraw() {

    // construct the view transformation
    //
    Vector p = context->get(GF_CA_POSN);
    Vector h = context->get(GF_CA_HEAD);
    Vector u = context->get(GF_CA_UP  );
    Matrix view = ::view(p, h, u);

    if (d3dd) d3dd->SetTransform(D3DTS_VIEW, (D3DXMATRIX*)&view);

    // set global ambient light
    //
    Vector c = context->get(GF_LT_BKGD);

    if (d3dd){
		d3dd->SetRenderState(D3DRS_CULLMODE,3);
		d3dd->SetRenderState(D3DRS_AMBIENT, D3DCOLOR_COLORVALUE(c.x, c.y, c.z, 1));
	}
    // clear the backbuffer
    //
    if (d3dd) {
        d3dd->Clear(0, NULL, D3DCLEAR_TARGET | D3DCLEAR_ZBUFFER, 
         D3DCOLOR_XRGB(BGROUND_R, BGROUND_G, BGROUND_B), 1.0, 0);
        d3dd->BeginScene();
		
	
			
			for (DWORD i=0; i<numMaterials; i++)
{
   // Set the material and texture for this subset
	d3dd->SetMaterial(&meshMaterials[i]);
    d3dd->SetTexture(0,meshTextures[i]);
        
  // Draw the mesh subset
  mesh->DrawSubset( i );
}
			
    }
    

}

// alphaBlendOn turns on alphaBlending
//
void Display::alphaBlendOn() {

    // turn on alpha blending
    if (d3dd) d3dd->SetRenderState(D3DRS_ALPHABLENDENABLE, TRUE);
}

// alphaBlendOff turns off alphaBlending
//
void Display::alphaBlendOff() {

    // turn off alpha blending
    if (d3dd) d3dd->SetRenderState(D3DRS_ALPHABLENDENABLE, FALSE);
}

// beginSuperimpose prepares the display device for drawing of the HUD
//
void Display::beginSuperimpose() {

    // prepare to draw the hud
    //
	if (spriteManager_)
		// start the sprite manager
		spriteManager_->Begin(D3DXSPRITE_ALPHABLEND);
}

// draw draws a single frame of the scene and superimposes the hud
//
void Display::endDraw() {

	// finish drawing the HUD
    //
	if (spriteManager_)
    	spriteManager_->End();

    // present the backbuffer to the primary buffer
    //
    if (d3dd) {
        d3dd->EndScene();
        if (FAILED(d3dd->Present(NULL, NULL, NULL, NULL)))
            error(L"Display::40 Failed to flip backbuffer");
    }
}

// device returns the address of the display device itself
//
void* Display::device() const {

    return d3dd;
}

// spriteManager returns the address of the sprite manager
//
void* Display::spriteManager() const {

    return spriteManager_;
}

// resize resets the aspect ratio and recalculates the projection matrix
//
void Display::resize() {

	if (runInWindow) {
        int width  = context->get(GF_WN_WDTH);
        int height = context->get(GF_WN_HGHT);

		// reset the aspect ratio
		aspect = float(width) / height;
		// reset projection transformation
		setupProjection();
	}
}

// suspend prepares the display device for de-activation
//
void Display::suspend() {

    // detach the sprite manager from video memory
    if (spriteManager_)
        spriteManager_->OnLostDevice();
}

// release releases the interfaces to the display device
//
void Display::release() {

    suspend();

    // release the font manager
    if (spriteManager_) {
        spriteManager_->Release();
        spriteManager_ = NULL;
    }
	// release the display device
    if (d3dd) {
        d3dd->Release();
        d3dd = NULL;
    }

    if (lightIndex) {
        delete [] lightIndex;
        lightIndex = NULL;
    }
}

// destructor releases the display object along with the Interface to the
// reference object
//
Display::~Display() {

    release();

    if (d3d) {
        d3d->Release();
        d3d = NULL;
    }

    // mode dimensions memory
    if (modeDim) {
        delete [] modeDim;
        modeDim = NULL;
    }
}
