/* Object Implementation - Scene Component - Model Branch
 *
 * consists of Object base class
 *             Box class
 *             Grid class
 *
 * Object.cpp
 * fwk4gps version 1.0
 * gam666/dps901
 * August 30 2010
 * copyright (c) 2010 Chris Szalwinski 
 * distributed under TPL - see ../Licenses.txt
 */

#include "iScene.h"           // for the Scene Interface
#include "iGraphic.h"         // for the Graphic Interface
#include "iTexture.h"         // for the Texture Interface
#include "iSound.h"           // for the Sound Interface
#include "iLight.h"           // for the Light Interface
#include "iCamera.h"          // for the Camera Interface
#include "iText.h"            // for the Text Interface
#include "iUtilities.h"       // for error()

#include "Object.h"           // for Object, Box, Grid class declarations
#include "MathDeclarations.h" // for MODEL_Z_AXIS
#include "ModelSettings.h"    // for POINT_LIST, ...
#include "MathDefinitions.h"
#define PI 3.14159265
//-------------------------------- Object -------------------------------------
//
// The Object Primitive represents a single object in the Scene Component
//
// The Scene Coordinator manages the Object Instances
//
iScene* Object::coordinator = NULL;

// Clone create a clone of Object* src and returns the address to its interface
//
iObject* Clone(const iObject* src) {

    return src->clone();
}

// constructor creates a 3-D scene object that is modelled by 'nPrimitives'
// graphics primitives of type 'pType' with material reflectivity 'c'
//
Object::Object(Colour c) : material(c), texture(NULL), sound(NULL), light(NULL), 
 noTexts(0) {
    
    if (!coordinator)
        error(L"Object::00 Couldn\'t access the Scene Coordinator");
    else if(!coordinator->add(this))
        error(L"Object::01 Couldn\'t add the object to the Scene Coordinator");

    flags   = TEX_DEFAULT;

	for (int i = 0; i < MAX_RPTS; i++) {
		text[i] = NULL;
		data[i] = NULL;
	}
}

// copy constructor initializes instance pointers and calls the assignment
// operator
//
Object::Object(const Object& src) {

    if (!coordinator)
        error(L"Object::00 Couldn\'t access the Scene Coordinator");
    else if(!coordinator->add(this))
        error(L"Object::01 Couldn\'t add the object to the Scene Coordinator");

    *this   = src;
}
            
// assignment operator discards old data and creates a copy from
// src - does not copy the texture or the text items attached to 
// src - instead initializes them to NULL - does not copy the 
// Graphic attached to src but creates a clone for the current
// object
//
Object& Object::operator=(const Object& src) {

    if (this != &src) {
        *((Frame*)this) = src;
        material        = src.material;
        flags           = src.flags;
        texture         = src.texture;
        sound           = NULL;
        camera          = NULL;
        light           = NULL;
        noTexts         = 0;
	    for (int i = 0; i < MAX_RPTS; i++) {
		    text[i] = NULL;
		    data[i] = NULL;
	    }
    }

    return *this;
}

// attach attaches iText* t to the object along with a pointer to function d for
// generating the text string to be reported
//
void Object::attach(const iText* t, const wchar_t* (*d)(wchar_t* s, const iObject*)) {

	if (noTexts < MAX_RPTS) {
		text[noTexts] = t;
		data[noTexts] = d;
		noTexts++;
	}
}

// attach attaches iTexture *t to the object and sets the filtering flags
//
void Object::attach(const iTexture* t, unsigned filter) {

	texture = t;
	if (filter) flags = filter;
}

// attach attaches iSound* s to the object and vice versa
//
void Object::attach(const iSound* s) {

	sound = s;
	if (sound)
		(const_cast<iSound*>(sound))->attach(this);
    else
		(const_cast<iSound*>(sound))->attach((iFrame*)NULL);
}

// attach attaches iLight* l to the object and vice versa
//
void Object::attach(const iLight* l) {

	light = l;
	if (light)
		(const_cast<iLight*>(light))->attach(this);
    else
		(const_cast<iLight*>(light))->attach((iFrame*)NULL);
}

// attach attaches iCamera* c to the object and vice versa
//
void Object::attach(const iCamera* c) {

	camera = c;
	if (camera)
		(const_cast<iCamera*>(camera))->attach(this);
    else
		(const_cast<iCamera*>(camera))->attach((iFrame*)NULL);
}

// draw draws the object if its opacity matches the opacity request
//
void Object::draw(bool opacity) const { 
	
	if (opacity == (material.a == 1.0f)) 
		draw(texture, flags); 
}

// report sets the text strings to be reported for the object
//
void Object::report() const {

	wchar_t str[MAX_DESC + 1];
	for (int i = 0; i < noTexts; i++)
		if (text[i] && data[i])
			const_cast<iText*>(text[i])->set(data[i](str, this));
}		

// destructor removes the object from the Scene coordinator
//
Object::~Object() {

    if (coordinator)
        coordinator->remove(this);
    else
        error(L"Object::90 Couldn\'t access the Scene Coordinator");
}

//-------------------------------- Box -------------------------------------
//
// Box is an Object identifiable by two points - lower close left, and
// upper faraway right - and a colour
//
iObject* CreateBox(float minx, float miny, float minz, float maxx, 
 float maxy, float maxz, Colour c) {

	return new Box(minx, miny, minz, maxx, maxy, maxz, c);
}

// constructor builds the Box from its two extreme points one face at a time
//
Box::Box(float minx, float miny, float minz, float maxx, float maxy,
 float maxz, Colour c) : Object(c) {

    graphic = CreateGraphic(TRIANGLE_LIST, 12);

    float x = (minx + maxx) / 2, y = (miny + maxy) / 2, z = (minz + maxz) / 2;
    minx -= x;
    miny -= y;
    minz -= z;
    maxx -= x;
    maxy -= y;
    maxz -= z;
    Vector p1 = Vector(minx, miny, minz),
           p2 = Vector(minx, maxy, minz),
           p3 = Vector(maxx, maxy, minz),
           p4 = Vector(maxx, miny, minz),
           p5 = Vector(minx, miny, maxz),
           p6 = Vector(minx, maxy, maxz),
           p7 = Vector(maxx, maxy, maxz),
           p8 = Vector(maxx, miny, maxz);
    graphic->add(p1, p2, p3, p4, Vector(0, 0, -1)); // front
    graphic->add(p4, p3, p7, p8, Vector(1, 0,  0)); // right
    graphic->add(p8, p7, p6, p5, Vector(0, 0,  1)); // back
    graphic->add(p6, p2, p1, p5, Vector(-1, 0, 0)); // left
    graphic->add(p1, p4, p8, p5, Vector(0, -1, 0)); // bottom
    graphic->add(p2, p6, p7, p3, Vector(0, 1,  0)); // top
    translate(x, y, z);
}

// copy constructor clones the graphic representation
//
Box::Box(const Box& src) : Object(src) {

    if (src.graphic)
        graphic = src.graphic->clone();
    else
        graphic = NULL;
}
            
// assignment operator discards old data and creates a copy from
// src - does not copy the Graphic attached to src but creates a 
// clone for the current object
//
Box& Box::operator=(const Box& src) {

    if (this != &src) {
        *((Object*)this) = src;
        if (graphic)
            graphic->Delete();
        if (src.graphic)
            graphic = src.graphic->clone();
        else
            graphic = NULL;
    }

    return *this;
}

// draw draws the graphic representation of the Box
//
void Box::draw(const iTexture* t, unsigned f) const { 
    
    if (graphic) graphic->draw(this, t, f); 
} 

// suspend suspends the object
//
void Box::suspend() const { 
	
	if (graphic) 
		const_cast<iGraphic*>(graphic)->suspend(); 
}

// release releases the object
//
void Box::release() const { 
	
	if (graphic) 
		const_cast<iGraphic*>(graphic)->release(); 
}

// destructor deletes the graphic model
//
Box::~Box() {

    if (graphic)
        graphic->Delete();
}

//-------------------------------------Cylinder-----------------------------

iObject* CreateCylinder(float Radius, float Height, float Sides, Colour c) {

	return new Cylinder(Radius,Height,Sides,c);
}

// constructor builds the Box from its two extreme points one face at a time
//
Cylinder::Cylinder(float Radius, float Height, float Sides,Colour c ) : Object(c) {

    graphic = CreateGraphic(TRIANGLE_STRIP, 5*Sides);

	float Theta = 0; //Current Angle
	float Inc = 0; //Angular increment
	float x = 0; //x coord
	float y = 0; //y coord
	float z = 0; //z coord
	int i = 0;
	Vector n; //vertex normal
	Vector point;
	float  tu = 0;
	float tv = 0;
	float UStep = 0; //step between U coords
	Inc = 2 * PI / Sides; //where each side has two triangles
	Theta = 0 ;
	UStep = 1/Sides;
	int sign = -1;
	for (i = 1; i <= Sides; i++)
	{
		//calculate vertices
		x = Radius * cosf(sign *Theta);
		y = 0;
		z = Radius * sinf( sign* Theta);	
		tu = -(0.5 * cosf(sign * Theta) + 0.5);
        tv = 0.5 * sinf(sign * Theta) + 0.5;	
		graphic->add(x, 0, z, 0, -1,0,tu,tv);
		if(sign == 1)
			sign = -1;
		else if(sign == -1){
			sign = 1;
		Theta = Theta+ Inc;
		}
	}
	tu= 0;
	//Theta = 0;
	for (i = 0; i <=Sides; i++)
	{
		//Calculate Vertices
		x = Radius * cosf(Theta);
	    y = Height;
		z = Radius * sinf(Theta);
		point = Vector(x,0,z);	
		n = normal(point);	
		graphic->add(x, 0, z, n.x, n.y,n.z,tu,1);
		graphic->add(x, y, z, n.x, n.y,n.z,tu,0);

		Theta = Theta+ Inc;
		tu = tu + UStep;

	}
	//tu= 0;
	Theta = PI;
	sign = 1;
	for (i = 1; i <= Sides; i++)
	{
	//	//Calculate Vertices
		x = Radius * cosf(sign *Theta);
		y = Height;
		z = Radius * sinf( sign* Theta);	
		tu = 0.5 * cosf(sign * Theta) + 0.5;
        tv = 0.5 * sinf(sign * Theta) + 0.5;	
		graphic->add(x, y, z, 0, 1,0,tu,tv);
		if(sign == 1){
			sign = -1;
			Theta = Theta+ Inc;
		}
		else if(sign == -1){
			sign = 1;
		
		}
	}


}

// copy constructor clones the graphic representation
//
Cylinder::Cylinder(const Cylinder& src) : Object(src) {

    if (src.graphic)
        graphic = src.graphic->clone();
    else
        graphic = NULL;
}
            
// assignment operator discards old data and creates a copy from
// src - does not copy the Graphic attached to src but creates a 
// clone for the current object
//
Cylinder& Cylinder::operator=(const Cylinder& src) {

  if (this != &src) {
        *((Object*)this) = src;
        if (graphic)
            graphic->Delete();
        if (src.graphic)
            graphic = src.graphic->clone();
        else
            graphic = NULL;
    }

    return *this;
}

// draw draws the graphic representation of the Box
//
void Cylinder::draw(const iTexture* t, unsigned f) const { 
    
    if (graphic) graphic->draw(this, t, f); 
} 

// suspend suspends the object
//
void Cylinder::suspend() const { 
	
	if (graphic) 
		const_cast<iGraphic*>(graphic)->suspend(); 
}

// release releases the object
//
void Cylinder::release() const { 
	
	if (graphic) 
		const_cast<iGraphic*>(graphic)->release(); 
}

// destructor deletes the graphic model
//
Cylinder::~Cylinder() {

    if (graphic)
        graphic->Delete();
}

//-------------------------------------- Grid ------------------------------
//
// A Grid is a set of mutually perpendicular lines in a plane defined 
// parallel to the x-y plane
//
iObject* CreateGrid(int min, int y, int max, int n, Colour clr) {

	return new Grid(min, y, max, n, clr);
}

// constructor builds the Grid one line at a time
//
Grid::Grid(int min, int y, int max, int n, Colour clr) : Object(clr) {

    graphic = CreateGraphic(LINE_LIST, 2 * (n + 1));

    int i, inc = (max - min) / (n - 1);
    for (i = min; i <= max; i += inc) {
        graphic->add((float)min, (float)y, (float)i, 0, 1, 0);
        graphic->add((float)max, (float)y, (float)i, 0, 1, 0);
        graphic->add((float)i, (float)y, (float)min, 0, 1, 0);
        graphic->add((float)i, (float)y, (float)max, 0, 1, 0);
    }
}

// copy constructor clones the graphic representation
//
Grid::Grid(const Grid& src) : Object(src) {

    if (src.graphic)
        graphic = src.graphic->clone();
    else
        graphic = NULL;
}
            
// assignment operator discards old data and creates a copy from
// src - does not copy the Graphic attached to src but creates a 
// clone for the current object
//
Grid& Grid::operator=(const Grid& src) {

    if (this != &src) {
        *((Object*)this) = src;
        if (graphic)
            graphic->Delete();
        if (src.graphic)
            graphic = src.graphic->clone();
        else
            graphic = NULL;
    }

    return *this;
}

// draw draws the graphic representation
//
void Grid::draw(const iTexture* t, unsigned f) const { 
    
    if (graphic) graphic->draw(this, t, f); 
} 

// suspend suspends the graphic representation
//
void Grid::suspend() const { 
	
	if (graphic)
		const_cast<iGraphic*>(graphic)->suspend(); 
}

// release releases the graphic representation
//
void Grid::release() const { 
	
	if (graphic) 
		const_cast<iGraphic*>(graphic)->release(); 
}

// destructor deletes the graphic representation
//
Grid::~Grid() {

    if (graphic)
        graphic->Delete();
}

//-------------------------------- Utilities -----------------------------
//
// orient? returns a text string with the orientation of the local ? axis
// of iObject* object
//
const wchar_t* orientX(wchar_t* str, const iObject* object) {

	if (object) {
        Vector v = object->orientation('x');
		sprintf(str, (int)(100*v.x), (int)(100*v.y), (int)(100*v.z));
	}

	return str;
}

const wchar_t* orientY(wchar_t* str, const iObject* object) {

	if (object) {
		Vector v = object->orientation('y');
		sprintf(str, (int)(100*v.x), (int)(100*v.y), (int)(100*v.z));
	}

	return str;
}

const wchar_t* orientZ(wchar_t* str, const iObject* object) {

	if (object) {
		Vector v = object->orientation('z');
		sprintf(str, (int)(100*v.x), (int)(100*v.y), (int)(100*v.z));
	}

	return str;
}
