#ifndef _OBJECT_H_
#define _OBJECT_H_

/* Object Declarations - Scene Component - Model Branch
 *
 * consists of Object base class
 *             Box class
 *             Grid class
 *
 * Object.h
 * fwk4gps version 1.0
 * gam666/dps901
 * August 30 2010
 * copyright (c) 2010 Chris Szalwinski 
 * distributed under TPL - see ../Licenses.txt
 */

#include "iObject.h"       // for the Object Interface
#include "ModelSettings.h" // for MAX_RPTS

//-------------------------------- Object -------------------------------------
//
// The Object Primitive is an independent object in the Scene Component
//
class iScene;
enum  PrimitiveType;

class Object : public iObject {

	static iScene*  coordinator;    // points to the scene coordinator

	const iTexture* texture;        // points to the attached texture, if any
	const iSound*   sound;          // points to the attached sound, if any
    const iLight*   light;          // points to the attached light, if any
    const iCamera*  camera;         // points to the attached camera, if any
	Colour          material;       // material reflectivity
	unsigned        flags;          // texture sampling flags

	int             noTexts;        // number of reporting text strings
	const iText*    text[MAX_RPTS]; // points to the reporting text strings
	const wchar_t*  (*data[MAX_RPTS])(wchar_t* s, const iObject*);  
							        // points to the reporting functions

    void release();

  protected:
    Object(Colour c);
    virtual ~Object();

  public:
	static void isCoordinatedBy(iScene* scene) { coordinator = scene; }
    Object(const Object& o);            
    Object& operator=(const Object& o); 
    iObject* clone() const { return NULL; }
	// initialization functions
	void   attach(const iTexture* t, unsigned filter);
	void   attach(const iSound*);
	void   attach(const iLight*);
	void   attach(const iCamera*);
	void   attach(const iText* t, const wchar_t* (*d)(wchar_t* s, const iObject*));
	// execution functions
    virtual void draw(const iTexture*, unsigned) const = 0;
	Colour colour() const { return material; }
	void   draw(bool opacity) const;
	void   report() const;
	// termination functions 
	void   Delete() const { delete this; }
};

//-------------------------------- Box ----------------------------------------
//
// Box is an Object identifiable by two points - lower close left, and
// upper faraway right - plus a colour
//
class iGraphic;

class Box : public Object {

    iGraphic* graphic; // points to the graphic representation

  protected:
    Box(float minx, float miny, float minz, float maxx, float maxy,
     float maxz, Colour c);

  public:
	friend iObject* CreateBox(float minx, float miny, float minz, 
	 float maxx, float maxy, float maxz, Colour c);
    Box(const Box& o);            
    Box& operator=(const Box& o); 
    iObject* clone() const { return new Box(*this); }
    void     draw(const iTexture* t, unsigned f) const; 
	void     suspend() const;
    void     release() const;
	virtual  ~Box();
};

//-------------------------------Cylinder-------------------------------------

class iGraphic;

class Cylinder : public Object {

    iGraphic* graphic; // points to the graphic representation

  protected:
    Cylinder(float Radius, float Height, float Sides, Colour c);

  public:
	friend iObject* CreateCylinder(float Radius, float Height, float Sides, Colour c);
    Cylinder(const Cylinder& o);            
    Cylinder& operator=(const Cylinder& o); 
    iObject* clone() const { return new Cylinder(*this); }
    void     draw(const iTexture* t, unsigned f) const; 
	void     suspend() const;
    void     release() const;
	virtual  ~Cylinder();
};

//-------------------------------- Grid ---------------------------------------
//
// Grid is a set of mutually perpendicular lines in a plane parallel to the x-y
// plane
//
class Grid : public Object {

    iGraphic* graphic; // points to the graphic representation

  protected:
    Grid(int min, int y, int max, int n, Colour clr);
	virtual ~Grid();

  public:
	friend iObject* CreateGrid(int min, int y, int max, int n, Colour clr);
    Grid(const Grid& o);            
    Grid& operator=(const Grid& o); 
    iObject* clone() const { return new Grid(*this); }
    void  draw(const iTexture* t, unsigned f) const; 
	void  suspend() const;
    void  release() const;
};

#endif
